#define BOOST_BIND_GLOBAL_PLACEHOLDERS

#include <set>
#include <chrono>

#include <boost/bind/bind.hpp>
#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>

#include <fcf/except.hpp>
#include <fcf/net/socket.hpp>
#include <fcf/log.hpp>
#include <fcf/process.hpp>

#include "subprocess.hpp"

namespace application { 

  enum cntrl_package_type_e {
    CPM_EVENT,
    CPM_HANDLER_TYPE,
  };

namespace {
  inline std::string to_ctrl_package(boost::property_tree::ptree& a_ptree, cntrl_package_type_e a_event_type = CPM_EVENT){
    switch (a_event_type) {
      case CPM_EVENT: {
        boost::property_tree::ptree root;
        root.put("type", "event");
        root.add_child("data", a_ptree);
        std::string data;
        std::stringstream ss;
        boost::property_tree::write_json(ss, root, false);  
        data += std::to_string(ss.str().length());
        data += "\n";
        data += ss.str();
        return data;
      }
      case CPM_HANDLER_TYPE: {
        boost::property_tree::ptree root;
        root.put("type", "check_memory_leak");
        root.add_child("data", a_ptree);
        std::string data;
        std::stringstream ss;
        boost::property_tree::write_json(ss, root, false);  
        data += std::to_string(ss.str().length());
        data += "\n";
        data += ss.str();
        return data;
      }
    }
    return std::string();
  }
} // non namespace

subprocesses_t::subprocesses_t(size_t a_maxmem_leak_protection, size_t a_maxmem)
  : _maxmem_leak_protection(a_maxmem_leak_protection), _maxmem(a_maxmem), _stop(true) {

}

subprocesses_t::~subprocesses_t() {
  stop();
}

int subprocesses_t::append(const char* a_name,
                            std::vector<std::string> a_argv1,
                            std::vector<std::string> a_argv2, 
                            std::vector<std::pair<std::string, std::string> > a_envv, 
                            std::vector<std::string> a_node_options, 
                            const char* a_user,
                            const char* a_password,
                            int a_port1,
                            int a_port2,
                            int a_control_port1,
                            int a_control_port2,
                            int a_start_timeout,
                            int a_stop_timeout,
                            bool a_leak_protected) {
  _processes.push_back({a_name, a_argv1, a_envv, a_node_options, a_user,a_password, a_port1, a_control_port1, a_start_timeout, a_stop_timeout, (pid_t)0, 0, true, a_leak_protected});
  _processes.push_back({a_name, a_argv2, a_envv, a_node_options, a_user,a_password, a_port2, a_control_port2, a_start_timeout, a_stop_timeout, (pid_t)0, 0, false, a_leak_protected});
  return (int)_processes.size() - 1;
}

void subprocesses_t::run(){
  _stop   = false;

  for(int i = 0; i < _processes.size(); ++i) {
    if (i%2 == 0) {
      _restart_process(i, false);
    }
  }

  _restart_flags.resize(_processes.size() / 2);
  _observers.resize(_processes.size() / 2);
  for(size_t i = 0; i < (_processes.size() / 2); ++i){
    _observers[i] = std::thread(&subprocesses_t::_observer, this, i*2, i*2+1);
  }
}

void subprocesses_t::stop(){
  if (_stop)
    return;
  _stop = true;
  {
    std::unique_lock<std::mutex> lock(_mutex);
    _condition.notify_all();  
  }

  for(std::thread& thread : _observers){
    if (thread.joinable())
      thread.join();
  }

  for(int i = 0; i < _processes.size(); ++i)
    _stop_process(i, true);
}

int subprocesses_t::get_process(SUBPROCESS_LEAK_PROTECTED a_lp){
  int m = 0, c, result = -1;
  for(int i = 0; i < _processes.size(); ++i) {
    if (!_processes[i].active || !_processes[i].pid)
      continue;

    if (_processes[i].leak_protected && a_lp == SLP_NO)
      continue;

    if (!_processes[i].leak_protected && a_lp == SLP_YES)
      continue;

    c = _processes[i].counter.load();
    if (result == -1 || m > c) {
      result = i;
      m = c;
    }
  }

  ++_processes[result].counter;
  return result;
}

subprocesses_t::subprocesses_state_t&  subprocesses_t::get_process_info(int a_index){
  return _processes[a_index];
}

void subprocesses_t::release_process(int a_index){
  if (_processes[a_index].counter.load() <= 0) { // reserve protect
    _processes[a_index].counter = 0;
    return;
  }
  --_processes[a_index].counter;
}

bool subprocesses_t::is_active(int a_procNum) {
  return _processes[a_procNum].active;
}

void subprocesses_t::restart(){
  std::unique_lock<std::mutex> lock(_mutex);
  for(size_t i = 0; i < _restart_flags.size(); ++i)
    _restart_flags[i] = true;
}

request_processing_information_t subprocesses_t::get_request_processing_information(std::string a_path, std::string a_context) {
  request_processing_information_t rpi;
  int process_index = get_process(SLP_YES);
  try{
    boost::property_tree::ptree ptree;
    ptree.put("path", a_path);
    ptree.put("context", a_context);
    std::string data(to_ctrl_package(ptree, CPM_HANDLER_TYPE));
    fcf::net::socket_sptr_t socket(fcf::net::create_socket("127.0.0.1", _processes[process_index].control_port));
    socket->set_timeout(5);
    socket->connect();
    socket->send_all(&data[0], (int)data.length());
    char buffer[1024];
    memset(&buffer[0], 0, sizeof(buffer));
    socket->recv(&buffer[0], sizeof(buffer));

    const char* beg = &buffer[0];
    const char* end = 0;

    end = std::strchr(beg, ' ');
    rpi.memory_leak_protection = std::atoi(std::string(beg, end).c_str());

    beg = end + 1;
    end = std::strchr(beg, ' ');
    rpi.client_socket_timeout = std::atoi(std::string(beg, end).c_str());

    beg = end + 1;
    end = std::strchr(beg, ' ');
    rpi.max_request_timeout = std::atoi(std::string(beg, end).c_str());

    beg = end + 1;
    end = std::strchr(beg, ' ');
    rpi.server_socket_timeout = std::atoi(std::string(beg, end).c_str());

    beg = end + 1;
    end = std::strchr(beg, ' ');
    rpi.max_response_timeout = std::atoi(std::string(beg, end).c_str());

    release_process(process_index);
  } catch(std::exception){
    release_process(process_index);
  }
  return rpi;
}

void subprocesses_t::_stop_process(int a_index, bool a_hide_errors, bool a_force){ // no exceptions
  if (!_processes[a_index].pid){
    return;
  }
  pid_t pid          = _processes[a_index].pid;
  int   stop_timeout = !a_force ? _processes[a_index].stop_timeout : 5;
  int   control_port = _processes[a_index].control_port;

  std::chrono::system_clock::time_point start = std::chrono::system_clock::now();
  try {
    _send_close(control_port, stop_timeout);
    while(true){
      if (!fcf::check_pid(pid)){
        _processes[a_index].pid = 0;
        break;
      }
      int wsec = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start).count();
      if (wsec > stop_timeout)
        throw std::runtime_error("Process termination timeout exceeded");
      std::this_thread::sleep_for(std::chrono::milliseconds(50));
    }
    _processes[a_index].pid = 0;
  } catch(const std::exception& e){
    if (!a_hide_errors)
      fcf::err << "Invalid process termination: " << e.what();
    fcf::close_process(pid);
    _processes[a_index].pid = 0;
    return;
  }
}

void subprocesses_t::_restart_process(int a_index, bool a_restart){
  _stop_process(a_index, false);

  std::string cmdline = _processes[a_index].name + " ";

  for(size_t i = 0; i < _processes[a_index].node_options.size(); ++i)
    cmdline += _processes[a_index].node_options[i] + " ";

  for(size_t i = 0; i < _processes[a_index].argv.size(); ++i)
    cmdline += _processes[a_index].argv[i] + " ";

  if (a_restart)
    cmdline += "--restart ";

  fcf::log << "Start process: " << cmdline;

  std::vector<std::string> args;
  for(size_t i = 0; i < _processes[a_index].node_options.size(); ++i)
    args.push_back(_processes[a_index].node_options[i]);
  for(size_t i = 0; i < _processes[a_index].argv.size(); ++i)
    args.push_back(_processes[a_index].argv[i]);

  pid_t pid = fcf::create_process(_processes[a_index].name.c_str(),
                                  args, 
                                  _processes[a_index].envv, 
                                  _processes[a_index].user.c_str(),
                                  _processes[a_index].password.c_str());
  if (pid) {
    std::chrono::system_clock::time_point start = std::chrono::system_clock::now();
    try {
      while(true){
        try{
          fcf::net::socket_sptr_t ps(fcf::net::create_socket("127.0.0.1", _processes[a_index].port));
          ps->set_timeout(_processes[a_index].start_timeout);
          ps->connect();
          break;
        } catch (const std::exception&){
          std::this_thread::sleep_for(std::chrono::milliseconds(50));
        }

        int stat = 0;
        if (!fcf::check_pid(pid)) {
          std::this_thread::sleep_for(std::chrono::milliseconds(100));
          throw fcf::except::process_start(cmdline.c_str());
        }
        int wsec = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - start).count();
        if (_processes[a_index].start_timeout > 0 && wsec >= _processes[a_index].start_timeout){
          throw fcf::except::process_timeout(cmdline.c_str());
        }
      }
    } catch(const std::runtime_error& e) {
      fcf::close_process(pid);
      throw e;
    } catch(const std::exception& e) {
      fcf::close_process(pid);
      throw e;
    }

    std::this_thread::sleep_for(std::chrono::milliseconds(100));
    fcf::log << "Process started successfully (PID:" << pid << "): " << cmdline;
  } else {
    _processes[a_index].pid = pid;
    throw fcf::except::process_start(cmdline.c_str());
  }

  _processes[a_index].pid = pid;
  try {
    _send_environment(_processes[a_index].control_port, 5);
  } catch(const std::exception& e){
    fcf::err << "Failed send environment to process: " << e.what();
  }
}

void subprocesses_t::_observer(int a_first_procindex, int a_second_procindex) {
  
  int   current_procindex = a_first_procindex;
  int   reserve_procindex = a_second_procindex;
  int   first_iteration = true;
  bool  restart_error = false;

  {
    std::unique_lock<std::mutex> lock(_mutex);
    _condition.wait_for(lock, std::chrono::seconds(5));
    if (_stop)
      return;
  }

  while(true) {
    std::unique_lock<std::mutex> lock(_mutex);

    if (_stop)
      return;

    try {
      bool restart_flag = _restart_flags[a_first_procindex/2];
      lock.unlock();
      size_t memsize = fcf::get_process_memory_size(_processes[current_procindex].pid)/1000000;
      lock.lock();

      if (restart_flag)
        _restart_flags[a_first_procindex/2] = false;

      if (_stop)
        return;

      if (first_iteration)
        fcf::log << "Process memory size  (PID:" << _processes[current_procindex].pid << ") " << memsize << "MB";

      size_t maxmem = _processes[current_procindex].leak_protected ? _maxmem_leak_protection : _maxmem;

      bool switch_over = !memsize || 
                         (memsize > maxmem && maxmem) || 
                         !_processes[current_procindex].pid || 
                         restart_flag ||
                         restart_error;

      if (switch_over){
        lock.unlock();
          fcf::log << "Retart process PID:" << _processes[current_procindex].pid << ". Memory: " << memsize << "MB. Memory limit: " << maxmem << "MB";
          _restart_process(reserve_procindex, true);
          _processes[current_procindex].active = false;
          _processes[reserve_procindex].active = true;

          // TODO: Добавить срочную перезагрузку процесса по флагу restart_flag
          std::this_thread::sleep_for(std::chrono::milliseconds(100));
          std::chrono::system_clock::time_point starttm = std::chrono::system_clock::now();
          while(true) {
            while(!!_processes[current_procindex].counter.load()) {
              std::this_thread::sleep_for(std::chrono::milliseconds(50));
              if (_stop)
                return;
              int dsec = (int)std::chrono::duration_cast<std::chrono::seconds>(std::chrono::system_clock::now() - starttm).count();
              if (restart_flag && dsec > _processes[current_procindex].stop_timeout) {
                _stop_process(current_procindex, false, true);
                break;
              }
            }
            std::this_thread::sleep_for(std::chrono::milliseconds(100));
            if (!_processes[current_procindex].counter.load())
              break;
          }

          std::swap(current_procindex, reserve_procindex);
          _stop_process(reserve_procindex, false);

        lock.lock();
      }

      restart_error = false;
    } catch (std::exception& e) {
      if (!_stop)
        fcf::err << "Observer failed:" << e.what();
      restart_error = true;
    } catch (...) {
      if (!_stop)
        fcf::err << "Unknown Exception in process observer";
      restart_error = true;
    }

    if (!lock.owns_lock())
      lock.lock();

    _condition.wait_for(lock, std::chrono::seconds(1));


    first_iteration = false;
  }
}

void subprocesses_t::_send_close(int a_port, int a_timeout){
  boost::property_tree::ptree ptree;
  ptree.put("name", "exit");
  std::string data(to_ctrl_package(ptree));
  fcf::net::socket_sptr_t socket(fcf::net::create_socket("127.0.0.1", a_port));
  socket->set_timeout(a_timeout);
  socket->connect();
  socket->send_all(&data[0], (int)data.length());
}

void subprocesses_t::_send_environment(int a_port, int a_timeout){
  boost::property_tree::ptree ptree;
  ptree.put("name", "environment");
  boost::property_tree::ptree ptree_processes;
  for(auto proc_conf : _processes) {
    boost::property_tree::ptree ptree_process;
    ptree_process.put("host", "localhost");
    ptree_process.put("port", proc_conf.control_port);
    ptree_processes.push_back(boost::property_tree::ptree::value_type("", ptree_process));
  }
  ptree.put_child("processes", ptree_processes);
  std::string data = to_ctrl_package(ptree);

  fcf::net::socket_sptr_t socket(fcf::net::create_socket("127.0.0.1", a_port));
  socket->set_timeout(a_timeout);
  socket->connect();
  socket->send_all(&data[0], (int)data.length());
  char buffer[1024];
  socket->recv(&buffer[0], sizeof(buffer));
}

}
